/* 
 * Desc:    Bigbob code for controlling a robot which moves in a maze.
 * Author:  Lucie Matusova, xmatus21
 * Date:    06.02.2013
 */

 #define DEBUG 

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <libplayerc++/playerc++.h>
#include <libplayerinterface/player.h>

#include <fcntl.h>     
#include <errno.h>
#include <termios.h>

#include <iostream>
#include <string>
#include <vector>
#include <dirent.h>

using namespace std;
using namespace PlayerCc;

struct robot_t {
  long int forwardSpeedX;
  long int forwardSpeedY;

  double DA;
  double DB;
  double DC;
  double DU;
  double DD;
  double DL;
  double DR;

} bigbob;


void print_help() {
  cout << "To start simulation please type: ./bigbob X Y D W" << endl;
  cout << "X - map width in metres (e.g. 100)" << endl;
  cout << "Y - map height in metres (e.g. 100)" << endl;
  cout << "D - distance (in metres) on which robot can see objects with his senzors (e.g. 100)" << endl;
  cout << "W - bit width of values sent to FITkits (defaults to 256)" << endl << endl;
}

void AvoidObstacles(robot_t *bigbob, SonarProxy &sp) {
  //will avoid obstacles closer than 40cm
  double avoidDistance = 0.4;

  // right sonar - sp[0]
  if(sp[0] < avoidDistance) {                  
    printf("cannot go to the right\n");
            
    if(sp[3] > avoidDistance) {
      bigbob->forwardSpeedX = -0.5;  // going left
      bigbob->forwardSpeedY = 0;
    }
    else if (sp[1] > avoidDistance) {
      bigbob->forwardSpeedX = 0;
      bigbob->forwardSpeedY = 0.5;  // going up
    }
    else if (sp[2] > avoidDistance) {
      bigbob->forwardSpeedX = 0;
      bigbob->forwardSpeedY = -0.5;  // going down
    }

    return;
  }
    
  // left sonar - sp[3]
  else if(sp[3] < avoidDistance) {
    printf("cannot go to the left\n");
           
    if (sp[2] > avoidDistance) {
      bigbob->forwardSpeedX = 0;
      bigbob->forwardSpeedY = -0.5;  // going down
    }
    else if(sp[0] > avoidDistance) {
      bigbob->forwardSpeedX = 0.5;  // going right
      bigbob->forwardSpeedY = 0;
    }
    else if (sp[1] > avoidDistance) {
      bigbob->forwardSpeedX = 0;
      bigbob->forwardSpeedY = 0.5;  // going up
    }            
      
    return;
  }
    
  // top sonar - sp[1]
  else if(sp[1] < avoidDistance)
  {
    printf("cannot go up\n");
            
    if (sp[3] > avoidDistance) {
      bigbob->forwardSpeedX = 0;
      bigbob->forwardSpeedY = -0.5;  // going down
      printf("going down\n");
    }
    else if(sp[0] > avoidDistance) {
      bigbob->forwardSpeedX = 0.5;  // going right
      bigbob->forwardSpeedY = 0;
      printf("going right\n");              
    }
    else if (sp[2] > avoidDistance) {
      bigbob->forwardSpeedX = 0;
      bigbob->forwardSpeedY = -0.5;  // going left
      printf("going left\n");
    }             
  
    return;
  }      
      
  // bottom sonar - sp[2]
  else if(sp[2] < avoidDistance) {
    printf("cannot go down\n");
              
    if (sp[1] > avoidDistance) {
      bigbob->forwardSpeedX = 0;
      bigbob->forwardSpeedY = 0.5;  // going up
    }                   
    else if(sp[0] > avoidDistance) {
      bigbob->forwardSpeedX = 0.5;  // going right
      bigbob->forwardSpeedY = 0;
    }       
    else if (sp[3] > avoidDistance) {
      bigbob->forwardSpeedX = 0;
      bigbob->forwardSpeedY = -0.5;  // going down
    }

    return;
  }
      
  return; //do nothing
}

/*
 * Opens the serial connection with FITkit
 */
int open_connection (const char *usb) {
  char txBuffer[10];
  int fd;
  struct termios tty_attributes;
   
  if ((fd = open(usb,O_RDWR|O_NOCTTY|O_NONBLOCK))<0) {
	  fprintf (stderr,"Open error on %s\n", strerror(errno));
	  exit(EXIT_FAILURE);
  } 
  else {
	  tcgetattr(fd,&tty_attributes);
     
	  // c_cflag
	  // Enable receiver
	  tty_attributes.c_cflag |= CREAD;		
     
	  // 8 data bit
	  tty_attributes.c_cflag |= CS8;			
    
	  // c_iflag
	  // Ignore framing errors and parity errors. 
	  tty_attributes.c_iflag |= IGNPAR;		
     
	  // c_lflag
	  // DISABLE canonical mode.
	  // Disables the special characters EOF, EOL, EOL2, 
	  // ERASE, KILL, LNEXT, REPRINT, STATUS, and WERASE, and buffers by lines.
     
	  // DISABLE this: Echo input characters.
	  tty_attributes.c_lflag &= ~(ICANON); 	 
    
	  tty_attributes.c_lflag &= ~(ECHO);		
     
	  // DISABLE this: If ICANON is also set, the ERASE character erases the preceding input 	
	  // character, and WERASE erases the preceding word.
	  tty_attributes.c_lflag &= ~(ECHOE);		
     
	  // DISABLE this: When any of the characters INTR, QUIT, SUSP, or DSUSP are received, generate the corresponding signal.	
	  tty_attributes.c_lflag &= ~(ISIG);		
     
	  // Minimum number of characters for non-canonical read.
	  tty_attributes.c_cc[VMIN]=1;			
     
	  // Timeout in deciseconds for non-canonical read.
	  tty_attributes.c_cc[VTIME]=0;			
     
	  // Set the baud rate
	  cfsetospeed(&tty_attributes,B460800);		
	  cfsetispeed(&tty_attributes,B460800);
     
	  tcsetattr(fd, TCSANOW, &tty_attributes);
  }	
   
  // ! close(fd);	
  return fd;
}

/*
 * Catching answer from FITkit's USB
 */
bool recieve_answer(int fd) {
  char buf[2000];

  // Read a char
  if (read(fd,&buf,1999) == -1) {
	  //cout << "Nothing new under the sun." << endl;
  } else {   
    string str = buf;

    if (str.find("YES") != string::npos) {
      // cout << "ANO!" << endl;
	    return true;
	  }
  }

  return false;
}

/*
 * Writes a connection request to given file descriptor
 */
bool send_request (int fd, char *req) {
  // write to port
  for (int i = 0; i <2; i++) {	
  	if (write(fd, req, strlen(req)) ==-1 ) {
      printf( "Send error: %s\n", strerror( errno ) );
    } 
    else {
      sleep(2);
    }
  }     
     
  if (!recieve_answer(fd)) {      
    return false;
  }

  return true;             
}
/*
 * Goes through /dev/, looks for ttyUSB interfaces 
 *  and tries to open connection with FITkits
 */
void usb_discovery(int *pc_to_fitkit, int *fitkit_to_pc) {
  vector<string> ports;
  vector<string>::iterator it;
  int fd;
  DIR *dir;
  struct dirent *ent;
  bool usb_in = false;
  bool usb_out = false;

  // open /dev/ and get its contents
  dir = opendir ("/dev/");
  if (dir != NULL) {
    string port;
    string dev = "/dev/";
       
    /* save all ttyUSBx port names to vector */
    while ((ent = readdir (dir)) != NULL) {
      port = ent->d_name;
      if (port.substr(0,6).compare("ttyUSB") == 0) {
        ports.push_back(dev + port);
      }
    }
    closedir (dir);
  } 
  else {
    /* could not open directory */
    fprintf (stderr,"Could not open /dev/.");
    return;
  }
      
  // send a connection request message to USB ports
  for (it = ports.begin(); it != ports.end(); ++it) {
    // open port and set FITkit connection details
    if ((fd = open_connection(it->c_str()))<0) {
      // ignore errors when trying to open ports
	    continue;
    } 
    else {                  
      // calling FITkit->PC first
      //  (too many data going through and it is not possible to catch an answer)
      char in[] = "IN\n";
      if (send_request(fd, in)) {
        *fitkit_to_pc = fd;
        usb_in = true;
        cout << "Connection with FITkit on the way to PC   established on port " << *it << "." << endl ;
        continue;
      }

      sleep(1);
                                                
      char out[] = "OUT\n";
      if (send_request(fd, out)) {
        *pc_to_fitkit = fd;
        usb_out = true;
        cout << "Connection with FITkit on the way to FPGA established on port " << *it << "." << endl ;
        continue;
      }
                               
      // if it gets here, it is not IN nor OUT, close the descriptor
      close(fd);
    }                  
  }
}


void print_distances(robot_t bigbob) {
  cout << "right:  " << bigbob.DR << endl;
  cout << "left:   " << bigbob.DL << endl;
  cout << "top:    " << bigbob.DU << endl;
  cout << "bottom: " << bigbob.DD << endl;

  cout << "distance from A: " << bigbob.DA << endl;
  cout << "distance from B: " << bigbob.DB << endl;
  cout << "distance from C: " << bigbob.DC << endl;
}


/* Distances in centimetres */
void compute_distances(player_pose3d_t pos, int A, int B, int C, robot_t *bigbob) {
  bigbob->DA = sqrt((A - pos.py)*(A - pos.py) + pos.px*pos.px);
  bigbob->DB = sqrt(pos.px*pos.px + pos.py*pos.py);
  bigbob->DC = sqrt(pos.py*pos.py + (C - pos.px)*(C - pos.px));
}

void convert_distances(robot_t *bigbob, int range, int bit_width, int height, int width) {
  bigbob->DL = bigbob->DL/range * bit_width;
  bigbob->DR = bigbob->DR/range * bit_width;
  bigbob->DU = bigbob->DU/range * bit_width;
  bigbob->DD = bigbob->DD/range * bit_width;
  
  double diagonal = sqrt(width*width + height*height);

  bigbob->DA = bigbob->DA/diagonal * bit_width;
  bigbob->DB = bigbob->DB/diagonal * bit_width;
  bigbob->DC = bigbob->DC/diagonal * bit_width;
}

/* 
 * Sending given sensoric data over USB to FITkit 
 */
void send(robot_t bigbob, int fd) {
  unsigned char buf[5];
  char value [2];
  unsigned int n;

  // distance from A
  n = sprintf(value, "%.2X", (unsigned int) floor(0.5 + bigbob.DA)); // rounded value
  buf[0] = 'D';
  buf[1] = 'A';
  buf[2] = value[0];
  buf[3] = value[1];
  buf[4] = '\n';
  write(fd, &buf, 5);
  cout << buf << endl;

  // distance from B
  n = sprintf(value, "%.2X", (unsigned int) floor(0.5 + bigbob.DB)); // rounded value
  buf[0] = 'D';
  buf[1] = 'B';
  buf[2] = value[0];
  buf[3] = value[1];
  buf[4] = '\n';
  write(fd, &buf, 5);
  cout << buf << endl;

  // distance from C  
  n = sprintf(value, "%.2X", (unsigned int) floor(0.5 + bigbob.DC)); // rounded value
  buf[0] = 'D';
  buf[1] = 'C';
  buf[2] = value[0];
  buf[3] = value[1];
  buf[4] = '\n';
  write(fd, &buf, 5);
  cout << buf << endl;
  
  // ----------
  // distance from the left
  n = sprintf(value, "%.2X", (unsigned int) floor(0.5 + bigbob.DL)); // rounded value
  buf[0] = 'D';
  buf[1] = 'L';
  buf[2] = value[0];
  buf[3] = value[1];
  buf[4] = '\n';
  write(fd, &buf, 5);
  cout << buf << endl;  
  
  // ----------
  // distance from the right
  n = sprintf(value, "%.2X", (unsigned int) floor(0.5 + bigbob.DR)); // rounded value
  buf[0] = 'D';
  buf[1] = 'R';
  buf[2] = value[0];
  buf[3] = value[1];
  buf[4] = '\n';
  write(fd, &buf, 5);
  cout << buf << endl;
  
  // ----------
  // distance from the top
  n = sprintf(value, "%.2X", (unsigned int) floor(0.5 + bigbob.DU)); // rounded value
  buf[0] = 'D';
  buf[1] = 'U';
  buf[2] = value[0];
  buf[3] = value[1];
  buf[4] = '\n';
  write(fd, &buf, 5);
  cout << buf << endl;

  // ----------
  // distance from the bottom
  n = sprintf(value, "%.2X", (unsigned int) floor(0.5 + bigbob.DD)); // rounded value
  buf[0] = 'D';
  buf[1] = 'D';
  buf[2] = value[0];
  buf[3] = value[1];
  buf[4] = '\n';
  write(fd, &buf, 5);
  cout << buf << endl;  
  
  /*
  strcpy(buf, "HELP\n");
	if (write(fd, &buf, 5) ==-1 )
          printf( "Send error: %s\n", strerror( errno ) );
  */
	
	return;
}

/*
 * Conversion of string hexa number to double.
 */
double htod(unsigned hexa, int pos) {
  double number;
  double zaklad;

  zaklad = pow(16, pos);

  if (hexa >= '0' and hexa <= '9') {
    number = (hexa - '0') * zaklad;
  } else {
    number = (hexa - 55) * zaklad;
  }

  return number;
}

/*
 * Reading data from FITkit's USB
 */
void get_results(int fd, robot_t *bigbob) {
  char rxBuffer[10000];
  int i, j;
  unsigned char first = 0;
  unsigned char second = 0;
  long int value;

  // Read a char
  if ((i = read(fd,&rxBuffer,9999))==-1) {
    printf("Nic noveho.\n");
    //printf("Recieve error: %s\n", strerror(errno));
  } else {

    #ifdef DEBUG   
      cout << rxBuffer << endl;
    #endif

    for (j=0; j < i; j++){
      bool one = false;

      if (rxBuffer[j]=='8' && (rxBuffer[j+1]=='1' || rxBuffer[j+1]=='2')) {
        #ifdef DEBUG   
          cout << "buffer: " << rxBuffer[j] << rxBuffer[j+1]<< rxBuffer[j+2]<< rxBuffer[j+3] << rxBuffer[j+4] << rxBuffer[j+5] << " ~ ";
        #endif

        // prvni cislice
        if (rxBuffer[j+3] >= '0' && rxBuffer[j+3] <= '9')
          first = rxBuffer[j+3];
        else if (rxBuffer[j+3] >= 'A' && rxBuffer[j+3] <= 'F')
          first = rxBuffer[j+3];

        // druha cislice - je tam vubec?
        if (rxBuffer[j+4] >= '0' && rxBuffer[j+4] <= '9')
          second = rxBuffer[j+4];
        else if (rxBuffer[j+4] >= 'A' && rxBuffer[j+4] <= 'F')
          second = rxBuffer[j+4];
        else {
          // hodnota mela jen jednu cislici
          one = true;
        }

/*
        if (one) {
          value = htod(first,0);
          //printf("value: %c, adr: %c%c\n", first, rxBuffer[j], rxBuffer[j+1]);
        } else {
          value = htod(first,1) + htod(second, 0);
          //printf("value: %c%c, adr: %c%c\n", first, second, rxBuffer[j], rxBuffer[j+1]);
        }
*/

        char* end;
        value = strtol(&rxBuffer[j+3], &end, 16);
        // if (*end) { cout << "error converting at: " << end; }
        if (value >= (1L << 7))
            value -= (1L << 8);

        switch (rxBuffer[j+1]) {
          case '1':
            bigbob->forwardSpeedX = value;

						#ifdef DEBUG 
              cout << "1. reg: " << value << ", buffer: " << rxBuffer[j+3] << rxBuffer[j+4] << endl;
            #endif

			      if (value != 0.0) {                   
			        //printf("first register: %f\n", value);
			      }
            break;
          
          case '2':
            bigbob->forwardSpeedY = value;

            #ifdef DEBUG
              cout << "2. reg: " << value << ", buffer: " << rxBuffer[j+3] << rxBuffer[j+4] << endl;
            #endif

			      if (value != 0.0) {
              ;
            }
            break;
        }
        
        // this if was in the switch above, it was a mistake I guess..?
        if (one) {
           j--;
        }
        j+=5;
          
      }
    }
  }

  #ifdef DEBUG
    cout << endl;
  #endif

  return;
}

int main(int argc, char *argv[]) {

  if (argc < 4) {
    print_help();
    return -1;
  }
      
  int height = atoi(argv[1]);
  double half_height = floor(height/2 + 0.5);
  int point_B = 0;
  int width = atoi(argv[2]);
  double half_width = floor(width/2 + 0.5);

  #ifdef DEBUG 
    cout << "A = [0;" << height  <<   "], ";
    cout << "B = ["   << point_B << ";0], ";
    cout << "C = ["   << width   << ";0]" << endl;
  #endif

  int range = atoi(argv[3]) * 100;  // the range of sensors in centimetres
  int bit_width = 255; 
      
  if (argc == 5) {
    bit_width = atoi(argv[4]) - 1;
  } 
      
  robot_t bigbob;
  bigbob.forwardSpeedX = 0;
  bigbob.forwardSpeedY = 0;

  int pc_to_fitkit = -1;
  int fitkit_to_pc = -1;
  usb_discovery(&pc_to_fitkit, &fitkit_to_pc);

  // have we found connected FITkits?
  if (pc_to_fitkit == -1 || fitkit_to_pc == -1) {
    fprintf (stderr, "Could not connect to FITkit(s).\n");
    exit(EXIT_FAILURE);
  }

  using namespace PlayerCc;
	try {
    PlayerClient    robot("localhost", 6665);

    Position2dProxy p2dProxy(&robot,0);
    SonarProxy      sonarProxy(&robot,0);
    SimulationProxy simProxy(&robot,0);
      
    srand(time(NULL));
	
    //enable motors
    p2dProxy.SetMotorEnable(1);

    //request geometries
    p2dProxy.RequestGeom();
    sonarProxy.RequestGeom();
    // laserProxy.RequestGeom();
	
    /* here so that laserProxy[90] doesn't segfault on first loop */
    robot.Read();

    while(true) {	
      // read from the proxies
      robot.Read();

	    player_pose3d_t pos;

      pos.px = p2dProxy.GetXPos() + half_width;
      pos.py = p2dProxy.GetYPos() + half_height;

      #ifdef DEBUG
        cout << endl << "[" << pos.px << "," << pos.py << "]" << endl;  
      #endif
    
      // distances in centimetres
      bigbob.DR = sonarProxy[0];
      bigbob.DU = sonarProxy[1];
      bigbob.DD = sonarProxy[2];
      bigbob.DL = sonarProxy[3];
  
      compute_distances(pos, height, point_B, width, &bigbob);
      // convert_distances(&bigbob, range, bit_width, height, width);
            
	    /* testing 
      bigbob.DA = 1;
      bigbob.DB = 2;
      bigbob.DC = 3;
      bigbob.DL = 4;
      bigbob.DR = 5;
      bigbob.DU = 6;
      bigbob.DD = 7;
      */

      #ifdef DEBUG
	      print_distances(bigbob);
      #endif

      send(bigbob, pc_to_fitkit);

      //receive new movement data
      get_results(fitkit_to_pc, &bigbob);
		
      //set motors
      // SetSpeed (double aXSpeed, double aYSpeed, double aYawSpeed)
      p2dProxy.SetSpeed(bigbob.forwardSpeedX, bigbob.forwardSpeedY, 0);
//      p2dProxy.SetSpeed(bigbob.forwardSpeedX, -8, 0);
            
     usleep(350000);
     //usleep(300000);
    }

    /*
    bigbob.DR = 4;
    bigbob.DU = 4;
    bigbob.DD = 4;
    bigbob.DL = 4;
    bigbob.DA = 4;
    bigbob.DB = 4;
    bigbob.DC = 4;
	  print_distances(bigbob);
    send(bigbob, pc_to_fitkit);
    sleep(4);
    //receive new movement data from the second fitkit
    get_results(fitkit_to_pc, &bigbob);
     */

  } catch (PlayerCc::PlayerError e) {
    std::cerr << e << std::endl;
    return -1;
  }

}
